#ifndef LIB_ERROR_HANDLER_MQH_INCLUDED
#define LIB_ERROR_HANDLER_MQH_INCLUDED
#property version "4.2";
/**********************************************************************************
 * Copyright (C) 2010-2022 Dominik Egert <info@freie-netze.de>
 *
 * This file is the Error handler library include file.
 *
 * MQLplus, including this file may not be copied and/or distributed 
 * without explecit permit by the author.
 * Author Dominik Egert / Freie Netze UG.
 **********************************************************************************
 *
 *  Version: 4.2
 *  State: public
 *
 *  File information
 *  ================
 *
 *  Use: resolve error codes to text
 *
 *      Intention:
 *
 *          Extend the basic MQL error handling functions to cover custom codes.
 *          Easy and portable usage.
 *
 *          The basic idea is to extend and not replace the original implementation
 *          of the functions ResetLastError(), GetLastError(), SetUserError()
 *          All functions may be used paralell and interswitched, although the lib_err
 *          will loose its up to date state when intermixing calls with each other.
 *          This especially affects the comment function. Elsewise, you can simply resolve
 *          error codes by calling err_ResolveLastError(). This works with predefined 
 *          error codes as well as with user defined error codes.
 *
 *
 *      General usage
 *
 *          Three functions are supported:
 *              void err_SetUserError(const ushort err_code, const string ...)
 *              void err_ResetUserError()   
 *              const string err_ResolveLastError()
 *              const string err_ResolveLastError(int& error_code)
 *              const int    err_ResolveLastError(string& error_msg)
 *
 *          The functions err_SetUserError() simply replace the original functions from MQL. 
 *          Original and this extended version do not interfere with each other. Although, the 
 *          MQL version will not retrieve a value, if called after err_ResolveLastError().
 *
 *          The additional function is used to retrieve last error code and corresponding text.
 *          This works with all defined error codes, as well as predefined system codes.
 *
 *
 *      Defining own error codes
 *
 *          Error codes are split up into groupds, so first a unique group id must be defined.
 *          This ID must be unique within the same compilation project and may overlap if the 
 *          target binaries are different.
 *
 *          For error codes to be assigned convenient, it is needed to define 
 *          a set and a return code, the set code will have a prefix. Every error code gets 
 *          defined by helper macros.
 *
 *          All definitions must be inclosed within the beginning and ending macros.
 *
 *          Here is an example of how it will look like:
 *
 *              #define EA_ERRGROUP_CODE_EXAMPLE            0x01
 *
 *              // Define return codes
 *              #define EXMPL_SUCCESS                       LIB_ERROR_CODE(_EXMPL_SUCCESS)
 *              #define EXMPL_ERR_INVALID_VALUE             LIB_ERROR_CODE(_EXMPL_ERR_INVALID_VALUE)
 * 
 *              // Define message resolver
 *              LIB_ERR_REGISTER_RESOLVER_BEGIN(EA_ERRGROUP_CODE_EXAMPLE)
 *                  LIB_ERR_REGISTER_RESOLVER_MSGCODE(EXMPL_SUCCESS,              "EXMPL_SUCCESS",              "Operation success. All values could be calculated within given parameters.")
 *                  LIB_ERR_REGISTER_RESOLVER_MSGCODE(EXMPL_ERR_INVALID_VALUE,    "EXMPL_ERR_INVALID_VALUE",    "Operation aborted. Invalid function input.")
 *              LIB_ERR_REGISTER_RESOLVER_END(EA_ERRGROUP_CODE_EXAMPLE)
 * 
 *
 *          IMPORTANT!!
 *
 *              The fisrt defined code MUST be the success code for routing to be correct. 
 *              It will be, no matter what you define here, always treated as success code.
 *              The success code is also equal to the group id.
 *
 *
 *          Reserved codes
 *
 *              To use reserved error codes, which is nothing different than splitting up the available
 *              error groups into two types, reserved and user defined, the defining macros in the example 
 *              above get exchanged with 
 *
 *                  LIB_ERROR_RESERVED_CODE()
 *                  LIB_ERR_RESERVED_REGISTER_RESOLVER_BEGIN()
 *                  LIB_ERR_RESERVED_REGISTER_RESOLVER_END()
 *
 *              Everything else stays the same. This enables to use internal or predefined error codes and
 *              still be able to define new codes, not interfering with each other due to duplicates.
 *
 *
 *          This now can be used anywhere in the code by calling...
 *
 *              SetUserError(EXMPL_SUCCESS, "Some additional comment!");
 *              SetUserError(EXMPL_SUCCESS, DBG_STR_PERSIST("Some additional comment! Will persist into release build."));
 *
 *
 *          ... and anywhere else be retrieved by
 *
 *              const string error_msg = err_ResolveLastError();
 *
 *                  ... or ...
 *                  
 *              int error_code = NULL;
 *              const string error_msg = err_ResolveLastError(error_code);
 *              
 *                  ... or ...
 *                  
 *              string error_msg = NULL
 *              const int error_code = err_ResolveLastError(error_msg);
 *
 *
 *      Hints:
 *
 *          It should be noted, retrieving the last error by calling err_ResolveLastError() will also reset the error state.
 *          There are a limited amount of user defined error codes available.
 *
 *              Current limit: ushort -> 16 Bit -> 65536 possible error codes.
 *                  The error code is split up into upper byte and lower byte. The upper byte is used as group id 
 *                  while the lower byte is used as error code id.
 *                  The group size geometry and the amount of internal group ids can be set below by altering the
 *                  appropiate defines.
 *
 *          In debug mode, the code will produce exits as any group id shows to be invalid. It may give false alarms 
 *          in case group ids are not ascendending defined. The runtime version will ignore such errors and simply 
 *          failroute to another router, not resolve at all and respond with "unknown error" or it simply works.
 *          In case all definitions are proper set up and the order is wrong at compile time, this will still result 
 *          in a working message router.
 *          
 *
*/
#ifdef DBG_MSG_TRACE_FILE_LOADER
    DBG_MSG_TRACE_FILE_LOADER;

#endif



//*********************************************************************************************************************************************************/
// Global error handler configuration
//


/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Setup Lib Error (for standalone compilation)
//
//  These definitions should be set before including the library
//

// Configure for shared binary library import
//#define LIB_ERROR_IMPORT_SHARED_LIBRARY

// Configure also for debugging use of shared library
//#define LIB_ERROR_FORCE_SHARED_LIBRARY_IMPORT

// Configure for shared binary compile mode
//#define LIB_ERR_BINARY_LIBRARY

// User supplied OnError handler (unavailable in binary library mode)
//#define LIB_ERR_USER_ERROR_HANDLER

// Generate unified function macro substitution
//#define LIB_ERROR_API_FUNCTION_SHADOWING

// Since MQL5 build 3520, openCL error resolving is supported
//#define LIB_ERR_INCLUDE_OPENCL_ERROR_RESOLVER

// Use custom supplied error messages
//#define LIB_ERROR_DESCRIPTION_FROM_FILE

// 
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Define error group id 
//  This ID must be unique within
//  the given project.
//
//  Valid values for Group ID are
//  between 0x00 and 0xEF
//  0xF0 to 0xFF is reserved.
//
//  Error code categories
//      0x00            = Succes 
//      0x01 ... 0xDF   = Errors
//      0xE0 ... 0xEF   = Warnings
//      0xF0 ... 0xFF   = Handled Errors
//
//  Define the size each defined error router can handle. 
//  (0x0100 = 256 here, while 0x00 is always preoccupied with ERR_SUCCESS)
//
#define LIB_ERR_LEVEL_WARNING       0x00E0
#define LIB_ERR_LEVEL_HANDLED_ERROR 0x00F0
#define LIB_ERR_INTERNAL_GROUP_IDS  0xF000
#define LIB_ERR_GROUP_SIZE          0x0100

//
/////////////////////////////////////////////////////////////////////////////////////////////////////




/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Define error messages output format
//

// Define error output format 
#define LIB_ERR_OUTPUT_FORMAT(_err_code_, _code_name_, _msg_) ::StringFormat("(%i - %s) %s", _err_code_, _code_name_, _msg_)

//
/////////////////////////////////////////////////////////////////////////////////////////////////////


//
// END Global error handler configuration
//*********************************************************************************************************************************************************/





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Include lib error
//

// Include lib error definitions
#include <MQLplus/lib_error/lib_error_definitions.mqh>

//
/////////////////////////////////////////////////////////////////////////////////////////////////////




//*********************************************************************************************************************************************************/
// User defined OnError() handler 
//
#ifdef LIB_ERR_USER_ERROR_HANDLER

// Namespace handling
#ifndef __MQL4_COMPATIBILITY_CODE__
    namespace LIB_ERROR_LIBRARY_PREFIX { 

#endif


/////////////////////////////////////////////////////////
//
//  OnError() Function
//
//  Generic error handling function.
//  Some return codes are not of interest, so
//  they get sorted out here and  are
//  surpressed of being journaled.
//
const bool UserOnError(uint& err, const string prefix = NULL)
{
    // Local init
    const 	static 	bool   	exec_env    = (MQLInfoInteger(MQL_PROFILER) || MQLInfoInteger(MQL_FORWARD) || MQLInfoInteger(MQL_OPTIMIZATION) || MQLInfoInteger(MQL_FRAME_MODE)) && (!MQLInfoInteger(MQL_VISUAL_MODE) || !MQLInfoInteger(MQL_DEBUG));
                            err         = _LastError;
    const           bool    user_code   = ((err & 0xFFFF0000) != NULL);
    const           int     err_mask    = (user_code) ? 0x000000FF : 0x0000FFFF;

    // Success code, treat warnings as success
    if( ((err & err_mask) == ERR_SUCCESS)
     || ((user_code) && ((err & LIB_ERR_LEVEL_WARNING) == LIB_ERR_LEVEL_WARNING)) )
    {
		err_ResetUserErrorCode();
		return(false); 
    }

    // Handled error codes...
    else if((user_code) && ((err & LIB_ERR_LEVEL_HANDLED_ERROR) == LIB_ERR_LEVEL_HANDLED_ERROR))
    #ifdef LIB_ERR_LEVEL_WARNING_AS_ERROR
        // ...return as error
        { return(true); }

	#else
        // ...return as non error
        { return(false); }

	#endif
            
    // Skip output for certain runtime environments
    if(exec_env)
    { return(true); }

    // Default action to perform on error
    printf("%s", err_ResolveLastErrorCode(-1, prefix));

    // Return error state
    return(true);
}
//
/////////////////////////////////////////////////////////


//	End namespace handling
#ifndef __MQL4_COMPATIBILITY_CODE__
    }; // Namespace LIB_ERROR_LIBRARY_PREFIX

#endif


//	End UserOnError() definition
#endif
//
// END User defined OnError() handler 
//*********************************************************************************************************************************************************/
#endif // LIB_ERROR_HANDLER_MQH_INCLUDED
